#!/usr/bin/python

# This code is in the public domain.

# Step 3: as Step 1, but this time cluster the points retrieved to
# combine the different photometry, then show sketches of the SED
# and let users select objects for closer inspection.

try:
  input = raw_input
except NameError:
  pass

import pickle
import os
import sys

from astropy import coordinates
from astropy import units as u
from astropy import table
from matplotlib import pyplot as plt
import numpy as np
import pyvo

import vohelper


# for rough SED: map filter UCDs to representative wavelengths
# to do this better, we'd need more takeup of the photometry DM
UCD_TO_WL = {
  "phot.mag;em.opt.u": 3.5e-7,
  "phot.mag;em.opt.b": 4.5e-7,
  "phot.mag;em.opt.v": 5.5e-7,
  "phot.mag;em.opt.r": 6.75e-7,
  "phot.mag;em.opt.i": 8.75e-7,
  "phot.mag;em.ir.j": 1.25e-6,
  "phot.mag;em.ir.h": 1.75e-6,
  "phot.mag;em.ir.k": 2.2e-6,
  "phot.mag;em.ir.3-4um": 3.5e-6,
  "phot.mag;em.ir.4-8um": 6e-6,
  "phot.mag;em.ir.8-15um": 11.5e-6,
  "phot.mag;em.ir.15-30um": 22.5e-6,
}


QUERIES = [
  ("twomass", "http://dc.zah.uni-heidelberg.de/tap",
    """SELECT TOP 1000000 raj2000, dej2000, jmag, hmag, kmag 
        FROM twomass.data
        WHERE 1=CONTAINS(
                      POINT('ICRS', raj2000, dej2000),
                      CIRCLE('ICRS', {ra}, {dec}, {radius}))
        	AND Jmag<15"""),
  ("allwise", "http://tapvizier.u-strasbg.fr/TAPVizieR/tap",
    """SELECT raj2000, dej2000, w1mag, w2mag, w3mag, w4mag
        FROM "II/328/allwise"
        WHERE 1=CONTAINS(
                      POINT('ICRS', raj2000, dej2000),
                      CIRCLE('ICRS', {ra}, {dec}, {radius}))
          AND w1mag<14"""),
  ("sdss", "http://gea.esac.esa.int/tap-server/tap",
    """SELECT ra, dec, u_mag, g_mag, r_mag, i_mag, z_mag
        FROM gaiadr1.sdssdr9_original_valid
        WHERE 1=CONTAINS(
                      POINT('ICRS', ra, dec),
                      CIRCLE('ICRS', {ra}, {dec}, {radius}))
          AND i_mag<16"""),]


def work_around_vizast_bug(col):
  """fixes a non-interability problem between VizieR and astropy: arraysize=1
  has not meant 1-array on Vizier-TAP.

  This function makes arrays of such 1-arrays arrays of scalars.
  """
  if not np.isscalar(col[0]) and col[0].shape==(1,):
    return col.__class__(data=col[:,0], name=col.name, mask=col.mask[:,0],
      unit=col.unit, meta=col.meta)
  else:
    return col


def work_around_sdss_ucd_bug(name, ucd):
  """guesses better UCDs for SDSS' botched ones.
  """
  if ucd=="em.opt;phot.mag":
    return {
      "u_mag": "phot.mag;em.opt.u",
      "g_mag": "phot.mag;em.opt.b",
      "r_mag": "phot.mag;em.opt.r",
      "i_mag": "phot.mag;em.opt.i",
      "z_mag": "phot.mag;em.opt.i",
    }[name]
  return ucd


def get_tables(ra, dec, radius):
  """returns pairs of (short_name, result) for the queries defined.

  For experimentation, we cache the results here; to clear the cache,
  delete the file cache.pickle.
  """
  if os.path.exists("cache.pickle"):
    with open("cache.pickle", "rb") as f:
      return pickle.load(f)

  results = []
  for short_name, access_url, query in QUERIES:
    service = pyvo.dal.TAPService(access_url)
    results.append(
      (short_name, service.run_sync(query.format(**locals())).table))

  with open("cache.pickle", "wb") as f:
    pickle.dump(results, f)

  return results


def get_coordinates_for_table(table):
  """returns SkyCoord objects for an astropy table.

  This uses pos.eq.*; meta.main UCDs to know where to look.
  """
  ra_column = vohelper.get_name_for_ucd(
    "pos.eq.ra;meta.main", table)
  dec_column = vohelper.get_name_for_ucd(
    "pos.eq.dec;meta.main", table)

  # fix broken metadata (sigh)
  if table[ra_column].unit=="Angle[deg]":
    table[ra_column].unit = "deg"
  if table[dec_column].unit=="Angle[deg]":
    table[dec_column].unit = "deg"
  
  return coordinates.SkyCoord(
# WORKAROUND!
    work_around_vizast_bug(table[ra_column]), 
    work_around_vizast_bug(table[dec_column]))


def force_scalar(val):
  """returns val[0] if val is an array, val otherwise.

  Again, this is a workaround for a vizier-astropy battle.
  """
  if np.isscalar(val):
    return val
  else:
    return val[0]


def make_photo_cluster(rows):
  """makes a pair of (position, photopoint) from a list of database
  rows.
  """
  pos = [None, None]
  phots = []

  for row in rows:
    for index, col in enumerate(row):
      name = row.columns[index].name
# WORKAROUND!
      ucd = work_around_sdss_ucd_bug(name,
        row.columns[index].meta.get("ucd", "").lower())

      if ucd.startswith("phot.mag"):
        col = force_scalar(col)
        if ucd in UCD_TO_WL:
          phots.append((UCD_TO_WL[ucd], col))
      elif ucd=="pos.eq.dec;meta.main":
        pos[1] = force_scalar(col)
      elif ucd=="pos.eq.ra;meta.main":
        pos[0] = force_scalar(col)

  return tuple(pos), sorted(phots)


def make_seds(tables, clusters):
  """returns a sequence of (position, photopoints) from database tables
  and the custer result.

  We select columns based on UCDs.
  """
  seds = []
  for cluster in clusters:
    seds.append(
      make_photo_cluster([tables[table_ind][1][row_ind] 
        for table_ind, row_ind in cluster]))
  return seds


def select_seds(seds):
  selected = []

  for pos, phots in seds:
    to_plot = np.array(phots)
    plt.semilogx(to_plot[:,0], to_plot[:,1], '-')
    plt.ylim([min(to_plot[:,1]), max(to_plot[:,1])])
    plt.ylabel("Mag",fontsize=15)
    plt.xlabel("Wavelength", fontsize=15)
    plt.show(block=False)
    selection = input("s)elect SED, q)uit, enter for next? ")
    if selection=="q":
      break
    if selection=="s":
      selected.append(pos)
    plt.cla()

  return selected
  

def main():
  if len(sys.argv)!=4:
    ra, dec, radius = 130.8, 3.4, 0.3
  else:
    ra, dec, radius = [float(a) for a in sys.argv[1:]]

  tables = get_tables(ra, dec, radius)

  clusters = vohelper.compute_multi_join([
    get_coordinates_for_table(t) for name, t in tables],
      0.2*u.arcsec)
 
  seds = make_seds(tables, clusters)
  
  selected = np.array(select_seds(seds))

  if not len(selected):
    sys.exit("Nothing selected, nothing written.")

  t = table.Table()
  t.add_column(table.Column(name='ra', 
    data=selected[:,0], 
    unit=u.degree, 
    meta={"ucd": "pos.eq.ra;meta.main"}))
  t.add_column(table.Column(name='dec', 
    data=selected[:,1], 
    unit=u.degree, 
    meta={"ucd": "pos.eq.dec;meta.main"}))
  with open("selected_positions.vot", "wb") as f:
    t.write(output=f, format="votable")


if __name__=="__main__":
  main()
